%{ open Typing %}
%{ open Syntax %}
%{ open Preprocessing %}
%token <string> NONCE
%token <string> NONCE_STAR
%token <string> CONST
%token <string> SESSKEY
%token <string> KEY
%token <string> VAR
%token LEFT_PARAN
%token RIGHT_PARAN
%token DOT
%token COMMA
%right COMMA
%token COLON
%token EXCLAMATION
%token PIPE
%left PIPE
%token ZERO
%token NEW
%token NEWKEY
%token IN
%token OUT
%token LET
%token ELSE
%token IF
%token THEN
%token MATCH
%token WITH
%token LL
%token LH
%token LS
%token KEYTYPE
%token EQKEYTYPE
%token SESKEYTYPE
%token PUBKEYTYPE
%token VERKEYTYPE
%token EQUAL
%token EQUIV
%token ARROW
%token ENC
%token AENC
%token INL
%token INR
%token PK
%token VK
%token DEC
%token ADEC
%token SIGN
%token VERIFY
%token HASH
%token PROJ1
%token PROJ2
%token EOF
%token SEPARATOR
%token OR
%right OR
%token PLUS
%right PLUS
%token STAR
%right STAR
%token XL
%token XR
%token XLR
%token LEFT_BRACE
%token RIGHT_BRACE
%token UNDERSCORE

%nonassoc ITE
%nonassoc ELSE

%start <(Syntax.key_env * Syntax.process  * Syntax.process) option> top
%%

label:
  | LL {LL}
  | LH {LH}
  | LS {LS}

var:
  | x = VAR {x}

nonce :
  | n = NONCE {Nonce n}
  | n = NONCE_STAR {NonceStar n}

term:
  | LEFT_PARAN ; t = term RIGHT_PARAN {t}
  | n = nonce { VNonce (Error ($startpos,$endpos), n) }
  | x = var { VVar (Error ($startpos,$endpos), x) }
  | c = CONST { VConst (Error ($startpos,$endpos),c) }
  | sk = SESSKEY { VKey (Error ($startpos,$endpos),sk) }
  | k = KEY { VKey (Error ($startpos,$endpos),k) }
  | PK ; LEFT_PARAN ; t = term ; RIGHT_PARAN { VPKey (Error ($startpos,$endpos),t) }
  | VK ; LEFT_PARAN ; t = term ; RIGHT_PARAN { VVerKey (Error ($startpos,$endpos),t) }
  | ENC ; LEFT_PARAN ; t = term ; COMMA ; t2 = term ; RIGHT_PARAN  { VSEnc (Error ($startpos,$endpos),t, t2)}
  | SIGN ; LEFT_PARAN ; t = term ; COMMA ; t2 = term ; RIGHT_PARAN  { VSign (Error ($startpos,$endpos),t, t2)}
  | AENC ; LEFT_PARAN ; t = term ; COMMA ; t2 = term ; RIGHT_PARAN  { VAEnc (Error ($startpos,$endpos),t, t2)}
  | HASH ; LEFT_PARAN ; t = term ; RIGHT_PARAN  { VHash (Error ($startpos,$endpos),t)}
  | INL ; LEFT_PARAN ; t = term ; RIGHT_PARAN  { VInl (Error ($startpos,$endpos),t)}
  | INR ; LEFT_PARAN ; t = term ; RIGHT_PARAN  { VInr (Error ($startpos,$endpos),t)}
  | t1 = term ; COMMA ; t2 = term { VPair (Error ($startpos,$endpos),t1, t2) }

destructor : 
  | DEC ; LEFT_PARAN ; x = var ; COMMA ; t = term ; RIGHT_PARAN { DSDec (Error ($startpos,$endpos),x, t)}
  | ADEC ; LEFT_PARAN ; x = var ; COMMA ; t = term ; RIGHT_PARAN { DADec (Error ($startpos,$endpos),x, t)}
  | VERIFY ; LEFT_PARAN ; x = var ; COMMA ; t = term ; RIGHT_PARAN { DVer (Error ($startpos,$endpos),x, t)}
  | PROJ1 ; LEFT_PARAN ; x = var ; RIGHT_PARAN ; { DProj1 (Error ($startpos,$endpos),x)}
  | PROJ2 ; LEFT_PARAN ; x = var ; RIGHT_PARAN ; { DProj2 (Error ($startpos,$endpos),x)}

proc:
  | ZERO { PZero (Error ($startpos,$endpos)) }
  | LEFT_PARAN ; p = proc1 ; RIGHT_PARAN {p}
  | NEW; n = nonce; COLON; l = label ; DOT ; p = proc { PNew (Error ($startpos,$endpos), n, l, p)}
  | NEWKEY; k = KEY; COLON; t = typ ; DOT ; p = proc { PNewKey (Error ($startpos,$endpos), k, t, p)}
  | OUT ; LEFT_PARAN ; t = term ; RIGHT_PARAN ; DOT ; p = proc { POut (Error ($startpos,$endpos), t, p)}
  | EXCLAMATION ; p = proc { PRepl (Error ($startpos,$endpos), p)} 
  | IN ; LEFT_PARAN ; x = var ; RIGHT_PARAN ; DOT ; p = proc { PIn (Error ($startpos,$endpos), x, p)}
  | LET ; x = var ; EQUAL ; d = destructor ; IN ; p = proc ; ELSE ; q = proc { PLet (Error ($startpos,$endpos), x, d, p, q)}
  | LET ; x = var ; EQUAL ; d = destructor ; IN ; p = proc { PLet (Error ($startpos,$endpos), x, d, p, PZero(Error ($startpos,$endpos)))} %prec ITE
  | IF ; t1 = term ; EQUAL ; t2 = term ; THEN ; p = proc ; ELSE ; q = proc { PIf (Error ($startpos,$endpos), t1, t2, p, q)}
  | IF ; t1 = term ; EQUAL ; t2 = term ; THEN ; p = proc { PIf (Error ($startpos,$endpos), t1, t2, p, PZero(Error ($startpos,$endpos)))} %prec ITE
  | MATCH ; x = var ; WITH ; INL ; LEFT_PARAN ; y = var ; RIGHT_PARAN ; ARROW ; p = proc ; PIPE ; INR ; LEFT_PARAN ; z = var ; RIGHT_PARAN ; ARROW ; q = proc { PMatch (Error ($startpos,$endpos), x, y, p, z, q) }

proc1:
  | p = proc1 ; PIPE ; q = proc1 {PPar (Error ($startpos,$endpos),p,q)}
  | p = proc { p }

bikey:
  | LEFT_PARAN ; k1 = KEY ; COMMA ; k2 = KEY ; RIGHT_PARAN {(k1,k2)}
  | k=KEY {(k,k)}

typ:
  | l = label { TLabel l}
  | t = key_type {t}
  | t1 = typ ; PLUS ; t2 = typ {TSum (t1, t2)}
  | t1 = typ ; OR ; t2 = typ {TOr (t1, t2)}
  | t1 = typ ; STAR  ; t2 = typ {TPair (t1, t2)}
  | LEFT_BRACE; XL; EQUAL ; tl =  term ; COMMA ; XR ; EQUAL ; tr = term ; RIGHT_BRACE {TRefine (tl,tr)}
  | LEFT_BRACE; XLR; EQUAL ; t =  term ; RIGHT_BRACE {TRefine (t,t)}
  | LEFT_PARAN ; t = typ ; RIGHT_PARAN {t}
  | LEFT_PARAN ; t = typ ; RIGHT_PARAN ; UNDERSCORE ; k=key_type { TSEnc (t,k) }
  | LEFT_BRACE ; t = typ ; RIGHT_BRACE ; UNDERSCORE ; k=key_type { TAEnc (t,TPubKey k) }
  | PUBKEYTYPE ; LEFT_PARAN ; t=typ ; RIGHT_PARAN {TPubKey t}
  | VERKEYTYPE ; LEFT_PARAN ; t=typ ; RIGHT_PARAN {TVerKey t}

key_type :
  | KEYTYPE ; LEFT_PARAN ; l = label ; COMMA ; t = typ; RIGHT_PARAN { TKey (l,t) }
  | EQKEYTYPE ; LEFT_PARAN ; l = label ; COMMA ; t = typ; RIGHT_PARAN { TEqKey (l,t) }
  | SESKEYTYPE ; LEFT_PARAN ; l = label ; COMMA ; t = typ; RIGHT_PARAN { TSessKey (l,t) }

key_def : 
  | k=bikey ; COLON ; t = key_type { (k,(t,false)) }

key_env:
  | kd = key_def ; COMMA ; ke = key_env {let ((k1,k2),t) = kd in KeyEnv.add (k1,k2) t ke }
  | kd = key_def {let ((k1,k2),t) = kd in KeyEnv.singleton (k1,k2) t} 

top : 
  | EOF {None}
  | ke = key_env; SEPARATOR ; p = proc1 ; EQUIV ; q = proc1 ; EOF {Some (ke, p,q)}
  | ke = key_env; SEPARATOR ; p = proc1 ; EOF {Some (ke,p,p)}
  | p = proc1 ; EQUIV ; q = proc1 ; EOF {Some (KeyEnv.empty, p,q)}
  | p = proc1 ; EOF {Some (KeyEnv.empty, p,p)}
