(* The syntax of our calculus *)

(* Basic Term Types *)

type nonce  = 
  | Nonce of string
  | NonceStar of string

type const  = string

type var = string

type key = string

(* error type, for printing error messages + posisions *)
type error =
  | Error of (Lexing.position * Lexing.position)
  | ErTODO

(*
 * TODO cleanup 
(* Public Keys (generated from secret key) *)
type pkey = 
  | PKey of key

(* Verification Keys (generated from secret key) *)
type vkey = 
  | VerKey of key
*)

(* Terms *)
type term = 
  | VNonce of error * nonce
  | VConst of error * const
  | VVar of error * var
  | VKey of error * key
  | VPKey of error * term
  | VVerKey of error * term
  | VPair of error * term * term
  | VInl of error * term
  | VInr of error * term
  | VHash of error * term
  | VSEnc of error * term * term
  | VAEnc of error * term * term
  | VSign of error * term * term

(* Labels *)  
type label =
  | LL 
  | LH
  | LS

(* Types *)
type ty =
  | TLabel of label
  | TPair of ty * ty
  | TOr of ty * ty
  | TSum of ty * ty
  | TKey of label * ty
  | TSessKey of label * ty
  | TEqKey of label * ty
  | TPubKey of ty
  | TVerKey of ty
  | TRefine of term * term
  | TSEnc of ty * ty
  | TAEnc of ty * ty

(* Destructor applications *)
type dest = 
  | DSDec of error * var * term
  | DADec of error * var * term
  | DProj1 of error * var
  | DProj2 of error * var
  | DVer of error * var * term

(* Processes *)
type process =
  | PZero of error
  | PNew of error * nonce * label * process
  | PNewKey of error * key * ty * process
  | POut of error * term * process
  | PIn of error * var * process
  | PPar of error * process * process
  | PRepl of error * process
  | PLet of error * var * dest * process * process
  | PIf of error * term * term * process * process
  | PMatch of error * var * var * process * var * process

module VarEnv = Map.Make
  (struct
    type t = var
    let compare = String.compare
   end)

module KeyEnv = Map.Make
  (struct
    type t = key * key
    let compare (k1, k2) (k1',k2') = 
      let i = String.compare k1 k1' in 
      if i = 0 then String.compare k2 k2' else i
   end)


(* Environment Types *)
type nonce_env = nonce -> label option
type var_env = ty VarEnv.t
(* Key env includes a boolean flag that is true for keys under replication *)
type key_env = (ty * bool) KeyEnv.t

(* Constraint Types *)
type constr = term * term
type constr_set = constr list
type constr_set_env = constr list * var_env
type constr_set_set =  constr_set_env list 
